import math
import random
from typing import Iterator, Optional, TypeVar, Sequence, Sized

import torch
from torch import Generator
from torch.utils.data import Dataset, Sampler


TCovariant = TypeVar('TCovariant', covariant=True)


class RandomSubsetSequenceSampler(Sampler[int]):
    def __init__(self, data_source: Sized, num_samples: int, generator: Optional[Generator] = None) -> None:
        super().__init__(data_source)

        generator if generator else get_generator()
        dataset_size = len(data_source)

        assert num_samples <= dataset_size

        self.num_samples = num_samples
        self.indices = torch.randperm(dataset_size, generator=generator).tolist()
        self.start_index = 0

    def __len__(self) -> int:
        return self.num_samples

    def __iter__(self) -> Iterator[int]:
        indices = self.indices[self.start_index:self.start_index + self.num_samples]

        if len(indices) < self.num_samples:
            self.start_index = self.num_samples - len(indices)
            indices += self.indices[:self.start_index]
        else:
            self.start_index += self.num_samples

        yield from indices


class RandomSubset(Dataset[TCovariant], Sized):
    def __init__(self, dataset: Sequence[TCovariant], size: int) -> None:
        self.dataset = dataset
        self.size = min(size, len(dataset))

    def __len__(self) -> int:
        return self.size

    def __getitem__(self, index: int) -> TCovariant:
        remainder = len(self.dataset) % self.size
        group_size = len(self.dataset) // self.size
        group_size += 1 if index < remainder else 0

        index_subjacent = index + self.size * random.randrange(group_size)
        element = self.dataset[index_subjacent]

        return element


def get_num_steps(dataset_size: int, batch_size: int, frequency: int = 10) -> int:
    return max(1, math.floor(dataset_size / batch_size / frequency))


def get_generator() -> Generator:
    seed = int(torch.empty((), dtype=torch.int64).random_().item())
    generator = Generator()
    generator.manual_seed(seed)

    return generator



import logging
import os
from typing import Any

import mlflow
import mlflow.artifacts as artifacts
import torch
from pytorch_lightning import LightningModule

import sfu_torch_lib.io as io
import sfu_torch_lib.utils as utils

logger = logging.getLogger(__name__)


def checkpoint_exists(run_id: str, filename: str = 'last.ckpt') -> bool:
    return filename in (artifact.path for artifact in artifacts.list_artifacts(run_id=run_id))


def get_localized_checkpoint_path(
    run_id: str,
    filename: str = 'last.ckpt',
    overwrite: bool = True,
    cache: bool = False,
) -> str | None:
    if not checkpoint_exists(run_id, filename):
        return None

    artifact_uri = mlflow.get_run(run_id).info.artifact_uri

    assert isinstance(artifact_uri, str)

    checkpoint_path = io.localize_cached(
        os.path.join(artifact_uri, filename),
        f'{run_id}_{filename}',
        overwrite,
        cache,
    )

    return checkpoint_path


def get_resumable_checkpoint_path(
    run_id: str | None,
    run_id_pretrained: str | None,
    filename: str = 'last.ckpt',
    overwrite: bool = True,
    cache: bool = False,
) -> tuple[str | None, bool]:
    if run_id:
        checkpoint_path = get_localized_checkpoint_path(run_id, filename, overwrite, cache)

        if checkpoint_path:
            return checkpoint_path, False

    if run_id_pretrained:
        checkpoint_path = get_localized_checkpoint_path(run_id_pretrained, filename, overwrite, cache)

        if checkpoint_path:
            return checkpoint_path, True

    return None, True


def get_checkpoint(run_id: str, filename: str = 'last.ckpt') -> dict[str, Any] | None:
    checkpoint_path = get_localized_checkpoint_path(run_id, filename)

    if checkpoint_path is None:
        return None

    with io.open(checkpoint_path) as checkpoint_file:
        checkpoint = torch.load(checkpoint_file)

    return checkpoint


def load_model[T: LightningModule](
    run_id: str,
    module_class: type[T] | None = None,
    filename: str = 'last.ckpt',
    overwrite: bool = True,
    cache: bool = False,
    **kwargs,
) -> T:
    checkpoint_path = get_localized_checkpoint_path(run_id, filename, overwrite, cache)

    assert checkpoint_path

    if module_class is not None:
        module_class_asserted = module_class

    else:
        module_class_asserted = utils.get_run_class(run_id)
        assert isinstance(module_class_asserted, T.__class__)

    with io.open(checkpoint_path) as checkpoint_file:
        model = module_class_asserted.load_from_checkpoint(checkpoint_file, **kwargs)

    return model


def load_checkpoint_state(checkpoint_path: str, model: LightningModule, strict: bool = True) -> None:
    device = torch.device('cuda') if torch.cuda.device_count() else torch.device('cpu')

    with io.open(checkpoint_path) as checkpoint_file:
        checkpoint = torch.load(checkpoint_file, device, weights_only=False)

    model.load_state_dict(checkpoint['state_dict'], strict)


def load_run_state(run_id: str, model: LightningModule, filename: str = 'last.ckpt', strict: bool = True) -> None:
    checkpoint_path = get_localized_checkpoint_path(run_id, filename)

    assert checkpoint_path

    load_checkpoint_state(checkpoint_path, model, strict)



import bisect
import collections
import copy
import itertools
import math
from typing import List, Union, Iterable, Iterator, TypeVar, Sequence, Tuple, Optional

import numpy as np
from torch.utils.data.sampler import BatchSampler, Sampler


T = TypeVar('T')


class ImageDataset:
    def __len__(self) -> int:
        raise NotImplementedError

    def get_height_and_width(self, index: int) -> Tuple[int, int]:
        raise NotImplementedError


class GroupedBatchSampler(BatchSampler):
    def __init__(self, sampler: Union[Sampler[int], Iterable[int]], group_ids: List[int], batch_size: int) -> None:
        """
        Wraps another sampler to yield a mini-batch of indices. It enforces that the batch only contain elements from
        the same group. It also tries to provide mini-batches which follows an ordering which is as close as possible
        to the ordering from the original sampler.

        :param sampler: Base sampler.
        :param group_ids: If the sampler produces indices in range [0, N), `group_ids` must be a list of `N` ints which
        contains the group id of each sample. The group ids must be a continuous set of integers starting from 0,
        i.e. they must be in the range [0, num_groups).
        :param batch_size: Size of mini-batch.
        """
        super().__init__(sampler, batch_size, drop_last=True)
        self.group_ids = group_ids

    def __iter__(self) -> Iterator[List[int]]:
        buffer_per_group = collections.defaultdict(list)
        samples_per_group = collections.defaultdict(list)
        num_batches = 0

        for index in self.sampler:
            group_id = self.group_ids[index]

            buffer_per_group[group_id].append(index)
            samples_per_group[group_id].append(index)

            if len(buffer_per_group[group_id]) == self.batch_size:
                yield buffer_per_group[group_id]
                num_batches += 1
                del buffer_per_group[group_id]

            assert len(buffer_per_group[group_id]) < self.batch_size

        # now we have run out of elements that satisfy the group criteria, let's return the remaining
        # elements so that the size of the sampler is deterministic
        expected_num_batches = len(self)
        num_remaining = expected_num_batches - num_batches

        if num_remaining > 0:
            # for the remaining batches, take first the buffers with the largest number of elements
            for group_id, _ in sorted(buffer_per_group.items(), key=lambda x: len(x[1]), reverse=True):
                remaining = self.batch_size - len(buffer_per_group[group_id])
                samples_from_group_id = repeat_to_at_least(samples_per_group[group_id], remaining)
                buffer_per_group[group_id].extend(samples_from_group_id[:remaining])

                assert len(buffer_per_group[group_id]) == self.batch_size

                yield buffer_per_group[group_id]

                num_remaining -= 1

                if num_remaining == 0:
                    break

        assert num_remaining == 0


def repeat_to_at_least(sequence: Sequence[T], n: int) -> List[T]:
    repeat_times = math.ceil(n / len(sequence))

    repeated = list(itertools.chain.from_iterable(itertools.repeat(sequence, repeat_times)))

    return repeated


def compute_aspect_ratios(dataset: ImageDataset, indices: Optional[Sequence[int]] = None) -> List[float]:
    if indices is None:
        indices = range(len(dataset))

    aspect_ratios = [width / height for height, width in map(dataset.get_height_and_width, indices)]

    return aspect_ratios


def quantize(x: List[float], bins: list[float]) -> List[int]:
    bins = copy.deepcopy(bins)
    bins = sorted(bins)

    quantized = list(map(lambda y: bisect.bisect_right(bins, y), x))

    return quantized


def create_aspect_ratio_groups(dataset: ImageDataset, k: int = 0) -> list[int]:
    aspect_ratios = compute_aspect_ratios(dataset)

    bins = (2 ** np.linspace(-1, 1, 2 * k + 1)).tolist() if k > 0 else [1.0]
    groups = quantize(aspect_ratios, bins)  # type: ignore

    return groups



from __future__ import annotations

import functools
import importlib
import re
from types import ModuleType
from typing import Callable, Iterable, Iterator, Sequence

import mlflow
from pytorch_lightning import LightningModule
from torch.nn import ELU, Identity, LeakyReLU, Module, ReLU, Sigmoid, Softmax, Tanh


class AttributeDict[T]:
    def __init__(self, dictionary: dict[str, T]) -> None:
        self.dictionary = dictionary

        for key, value in self.items():
            setattr(self, key.lower().replace(' ', '_'), value)

    def __getattr__(self, name):
        return getattr(self, name)

    def __getitem__(self, key):
        return self.dictionary[key]

    def __repr__(self):
        return repr(self.dictionary)

    def __len__(self):
        return len(self.dictionary)

    def __contains__(self, item):
        return item in self.dictionary

    def __iter__(self):
        return iter(self.dictionary)

    def __or__(self, other: AttributeDict[T]) -> AttributeDict[T]:
        return AttributeDict(self.dictionary | other.dictionary)

    def update(self, other: AttributeDict[T]) -> AttributeDict[T]:
        return self | other

    def keys(self):
        return self.dictionary.keys()

    def values(self):
        return self.dictionary.values()

    def items(self):
        return self.dictionary.items()


def get_pairs[T](items: Iterable[T]) -> Iterator[tuple[T, T]]:
    first = None

    for index, item in enumerate(items):
        if index % 2 == 0:
            first = item
        else:
            yield first, item  # type: ignore


def create_selector[T](indices: int | Sequence[int]) -> Callable[[Sequence[T]], T | tuple[T, ...]]:
    def select(items: Sequence[T]) -> T | tuple[T, ...]:
        if isinstance(indices, int):
            return items[indices]

        return tuple(items[index] for index in indices)

    return select


def to_list[T](*args: Sequence[T] | T) -> list[T]:
    items: list[T] = []

    for argument in args:
        if isinstance(argument, Sequence):
            items += argument
        else:
            items.append(argument)

    return items


def to_bool(value: str | None) -> bool | None:
    if value is None:
        return None

    if value == '0' or value.lower() == 'false':
        return False

    return bool(value)


def to_bool_or_false(value: str | None) -> bool:
    return to_bool(value) or False


def get_activation(name: str) -> Module:
    return {
        'elu': ELU,
        'identity': Identity,
        'relu': ReLU,
        'leaky_relu': LeakyReLU,
        'sigmoid': Sigmoid,
        'softmax': Softmax,
        'tanh': Tanh,
    }[name]


def get_class(model_type: str, default_module: ModuleType | None = None) -> type:
    if '.' in model_type:
        module_name, class_name = model_type.rsplit('.', maxsplit=1)
        module = importlib.import_module(module_name)

    elif default_module is not None:
        module = default_module
        class_name = model_type

    else:
        raise ValueError('default_module not provided')

    model_class = getattr(module, class_name)

    return model_class


def get_run_class(
    run_id: str,
    default_module: ModuleType | None = None,
    key: str = 'model-type',
) -> LightningModule:
    model_type = mlflow.get_run(run_id).data.params[key]

    model_class = get_class(model_type, default_module)

    assert isinstance(model_class, LightningModule)

    return model_class


def get_type_or_run_class(
    model_type: str | None = None,
    run_id: str | None = None,
    default_module: ModuleType | None = None,
    key: str = 'model-type',
) -> LightningModule:
    if model_type is not None:
        model_class = get_class(model_type, default_module)
        assert isinstance(model_class, LightningModule)

    else:
        assert run_id is not None
        model_class = get_run_class(run_id, default_module, key)

    return model_class


def parse_map_sequence[T](string: str, sequence_type: Callable[[str], T]) -> dict[str, Sequence[T]]:
    substrings = re.split(r'(?<=]),', string)

    output = {}

    for substring in substrings:
        key, value = substring.split(':')

        output[key] = list(map(sequence_type, value.strip('[]').split(',')))

    return output


parse_map_ints = functools.partial(parse_map_sequence, sequence_type=int)
parse_map_floats = functools.partial(parse_map_sequence, sequence_type=float)
parse_map_strings = functools.partial(parse_map_sequence, sequence_type=str)
